import torch
import torch.nn as nn
import numpy as np

import torch
import torch.optim as optim
import torch.nn.functional as F

from layers import GraphAttentionLayer, SpGraphAttentionLayer


class graph_constructor(nn.Module):
    def __init__(self, nnodes, k, dim, device, alpha=3, static_feat=None):
        super(graph_constructor, self).__init__()
        self.nnodes = nnodes
        if static_feat is not None:
            xd = static_feat.shape[1]
            self.lin1 = nn.Linear(xd, dim)
            self.lin2 = nn.Linear(xd, dim)
        else:
            self.emb1 = nn.Embedding(nnodes, dim)
            self.emb2 = nn.Embedding(nnodes, dim)
            self.lin1 = nn.Linear(dim,dim)
            self.lin2 = nn.Linear(dim,dim)

        self.device = device
        self.k = k
        self.dim = dim
        self.alpha = alpha
        self.static_feat = static_feat

    def forward(self, idx):
        if self.static_feat is None:
            nodevec1 = self.emb1(idx.to(torch.int64))
            nodevec2 = self.emb2(idx.to(torch.int64))
        else:
            nodevec1 = self.static_feat[idx,:]
            nodevec2 = nodevec1

        nodevec1 = torch.tanh(self.alpha*self.lin1(nodevec1))
        nodevec2 = torch.tanh(self.alpha*self.lin2(nodevec2))

        a = torch.mm(nodevec1, nodevec2.transpose(1,0))-torch.mm(nodevec2, nodevec1.transpose(1,0))
        adj = F.relu(torch.tanh(self.alpha*a))
        mask = torch.zeros(idx.size(0), idx.size(0)).to(self.device)
        mask.fill_(float('0'))
        s1,t1 = adj.topk(self.k,1)
        mask.scatter_(1,t1,s1.fill_(1))
        adj = adj*mask
        return adj

    def fullA(self, idx):
        if self.static_feat is None:
            nodevec1 = self.emb1(idx)
            nodevec2 = self.emb2(idx)
        else:
            nodevec1 = self.static_feat[idx,:]
            nodevec2 = nodevec1

        nodevec1 = torch.tanh(self.alpha*self.lin1(nodevec1))
        nodevec2 = torch.tanh(self.alpha*self.lin2(nodevec2))

        a = torch.mm(nodevec1, nodevec2.transpose(1,0))-torch.mm(nodevec2, nodevec1.transpose(1,0))
        adj = F.relu(torch.tanh(self.alpha*a))
        return adj



class GAT(nn.Module):
    def __init__(self, nfeat, nhid, nclass, dropout, alpha, nheads, num_nodes, max_neigh, node_dim, device, add_feature_dim,
                 feature_size, addition_feature_size, output_dim, encoder_length, decoder_length, lstm_size):
        """Dense version of GAT."""
        super(GAT, self).__init__()
        self.gc = graph_constructor(num_nodes, max_neigh, node_dim, device)
        self.num_nodes = num_nodes
        self.idx = torch.arange(self.num_nodes).to(device)
        self.dropout = dropout

        self.attentions = [GraphAttentionLayer(nfeat, nhid, dropout=dropout, alpha=alpha, concat=True) for _ in range(nheads)]
        for i, attention in enumerate(self.attentions):
            self.add_module('attention_{}'.format(i), attention)

        self.out_att = GraphAttentionLayer(nhid * nheads, nclass, dropout=dropout, alpha=alpha, concat=False)
        self.final_out_1 = nn.Linear(num_nodes + add_feature_dim + lstm_size, lstm_size)
        self.final_out_2 = nn.Linear(lstm_size, 1)

        # 以下开始是LSTM部分
        self.encoder = nn.LSTM(input_size=feature_size + 1, hidden_size=lstm_size, num_layers=1, bidirectional=False,
                               batch_first=True, dropout=0.1)  # if batch_first==False: input_shape=[length,batch_size,embedding_size]
        # decoder的inputsize需要变化，加上2倍的LSTM_size，因为attention的维度就是这个
        self.decoder = nn.LSTM(input_size=addition_feature_size + lstm_size, hidden_size=lstm_size, num_layers=1,
                               batch_first=True, dropout=0.1)
        # attention的全连接
        self.attention_fc_1 = nn.Linear(2 * lstm_size, 2 * lstm_size)
        self.attention_fc_2 = nn.Linear(2 * lstm_size, 1)
        # 最后预测的FC的维度也需要修改

        self.fc_1 = nn.Linear(addition_feature_size +  lstm_size + lstm_size, 2 * lstm_size)
        self.fc_2 = nn.Linear(2 * lstm_size, output_dim)
        #self.training = training
        self.encoder_length = encoder_length
        self.decoder_length = decoder_length

    def attention_forward(self, dec_prev_hidden, enc_output):
        prev_dec_h = dec_prev_hidden[0].squeeze(dim=0).unsqueeze(1).repeat(1, self.encoder_length, 1)
        # prev_dec_h = dec_prev_hidden[0].repeat(1,7,1)
        atten_input = torch.cat([enc_output, prev_dec_h], dim=-1)  # batch*length*3*lstm_size
        attention_weights = self.attention_fc_2(F.relu(self.attention_fc_1(atten_input)))  # batch_size * length *1的权重矩阵
        attention_weights = F.softmax(attention_weights, dim=1)  # alpha: batch*length*1
        atten_output = torch.sum(attention_weights * enc_output, dim=1).unsqueeze(1)# bs *1 *(2*lstm_size)

        # 复制attention的维度
        atten_output = atten_output.repeat(1, self.decoder_length, 1)



        #dec_output, dec_hidden = self.decoder(dec_lstm_input, dec_prev_hidden)
        # dec_output: bs*1*lstm_size, dec_hidden [bs*1*lstm_size, bs*1*lstm_size]
        #
        return atten_output,  attention_weights

    def forward(self, x, idx, add_feature, lstm_input):
        if idx is None:
            adj = self.gc(self.idx)
        else:
            adj = self.gc(idx)

        x = F.dropout(x, self.dropout, training=self.training)
        x = torch.cat([att(x, adj) for att in self.attentions], dim=2)
        x = F.dropout(x, self.dropout, training=self.training)
        x = F.elu(self.out_att(x, adj))
        enc_output, enc_hidden = self.encoder(lstm_input)

        dec_prev_hidden = [enc_hidden[0][0].unsqueeze(0), enc_hidden[1][0].unsqueeze(0)]

        atten_output, atten_weights = self.attention_forward(dec_prev_hidden,
                                                                                     enc_output)




        x = self.final_out_1(torch.cat([x.squeeze(), add_feature.squeeze(), atten_output.squeeze()], dim=-1))
        x = self.final_out_2(F.leaky_relu(x))
        return x, adj










